from waivek.timer import Timer
from datetime import datetime, timezone
from zoneinfo import ZoneInfo
from waivek.ic import ic
from waivek.color import Code
# from .error import handler
timer = Timer()
import rich
console = rich.get_console()

# PEP 696, disallowing naive datetime objects via type annotations
#
#     [1] https://github.com/python/mypy/issues/10067 - I want to be able to declare "naive datetime" or "aware datetime" as types
#     [2] https://discuss.python.org/t/pep-696-type-defaults-for-typevarlikes/22569

# considerations:
#     
#     [1] datetime.now() is not timezone aware
#     [2] str(datetime) is not ISO 8601 compliant, e.g. `2021-08-31 12:34:56` instead of `2021-08-31T12:34:56`
#     [3] datetime.now(...) has microsecond precision, which is verbose for log files

# BLACKLIST: datetime.now(), str(datetime)

# convert : dt.astimezone(ZoneInfo("Asia/Kolkata"))
# epoch   : int(dt.timestamp())
# string  : dt.isoformat()
# is_naive: dt.tzinfo is None

# timezones:
#
#   timezone.utc
#   ZoneInfo("Asia/Kolkata")
#   ZoneInfo("UTC")
#
# to get the local timezone in a cross-platform way, you can use the tzlocal module

def print_datetime_flaws():
    ic(datetime.now())
    print("... now() is not timezone-aware\n")
    ic(str(datetime.now()))
    print("... str() missing 'T' and timezone or 'Z\n'")
    ic(datetime.now().isoformat())
    print("... isoformat() is verbose, includes microseconds, and missing timezone\n")
    ic(datetime.now(ZoneInfo("Asia/Kolkata")))
    print("... missing 'T'\n")
    ic(datetime.now(ZoneInfo("UTC")).isoformat())
    ic(datetime.now(ZoneInfo("Asia/Kolkata")).isoformat())
    print("... valid\n")

    print(".fromisoformat():")
    ic(datetime.fromisoformat(str(datetime.now())))
    ic(datetime.fromisoformat(str(datetime.now().isoformat())))
    ic(datetime.fromisoformat("2024-08-18 16:54:26.331658+00:00"))

    print("datetime.fromisoformat(2024-08-18 16:54:26.331658Z)" + (Code.LIGHTRED_EX + " ERROR"))


    


def get_now(zone: str): # for IST, use "Asia/Kolkata"
    return datetime.now(ZoneInfo(zone)).replace(microsecond=0)

def foo():
    dt = DateTimeTZ.now(ZoneInfo("Asia/Kolkata"))
    ic(dt.astimezone(ZoneInfo("UTC")))
    ic(int(dt.timestamp()))
    ic(dt.isoformat())

    epoch = 1722394690
    epoch.bit_count

    x = 1
    y = { "zero": 0 }
    print(x/y["zero"])

class DateTimeTZ(datetime):
    """
    A subclass of datetime.datetime that requires all instances to be timezone-aware.
    """
    def __new__(cls, *args, **kwargs):
        # Create the datetime instance
        dt = super().__new__(cls, *args, **kwargs)
        # Check if tzinfo is None or not
        if dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None:
            raise ValueError("Datetime object must be timezone-aware")
        return dt

    @classmethod
    def now(cls, tz): # pyright: ignore[reportIncompatibleMethodOverride]
        return super().now(tz=tz)

    @classmethod
    def from_datetime(cls, dt: datetime):
        return cls(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, dt.microsecond, dt.tzinfo)

if __name__ == "__main__":
    try:
        # foo()
        print_datetime_flaws()
    except Exception as e:
        # handler(e)
        console.print_exception(show_locals=True, extra_lines=0)





import os
import sys
import textwrap
from waivek.reltools import pathjoin

def get_width():
    try:
        return os.get_terminal_size().columns
    except:
        return 80

def bat(path):
    path = path if os.path.isabs(path) else pathjoin(sys._getframe(1), path)
    with open(path, 'r') as f:
        contents = f.readlines()
    lines = contents

    max_line_number = len(lines)
    line_number_width = len(str(max_line_number))
    gutter_padding = 2  # Padding on each side of the line number

    # Calculate the total width of the output
    terminal_width = get_width()
    gutter_width = line_number_width + 2 * gutter_padding + 1  # +1 for the vertical bar
    content_width = terminal_width - gutter_width - 1  # -1 for the space after the vertical bar

    # ANSI escape codes for dimming and resetting
    dim = "\033[2m"
    reset = "\033[0m"
    bold = "\033[1m"

    # Print the header
    print(f"{dim}{'─' * (gutter_width - 1)}┬{'─' * (content_width + 1)}{reset}")
    print(f"{dim}{'':{gutter_width - 1}}│{reset} File: {bold}{path}{reset}")
    print(f"{dim}{'─' * (gutter_width - 1)}┼{'─' * (content_width + 1)}{reset}")

    # Print the file contents with line numbers and wrapping
    for i, line in enumerate(lines, start=1):
        line = line.rstrip('\n')
        wrapped_lines = textwrap.wrap(line, width=content_width)
        for j, wrapped_line in enumerate(wrapped_lines):
            if j == 0:
                print(f"{dim}{i:>{line_number_width + gutter_padding}}{' ' * gutter_padding}│{reset} {wrapped_line}")
            else:
                print(f"{dim}{'':{line_number_width + gutter_padding}}{' ' * gutter_padding}│{reset} {wrapped_line}")

    # Print the footer
    print(f"{dim}{'─' * (gutter_width - 1)}┴{'─' * (content_width + 1)}{reset}")

if __name__ == "__main__":
    bat(sys.argv[1])


# run.vim: vert term python % tmp/lorem_ipsum.txt



import os
from waivek.nest.nested import greet_nested, new_addition
from waivek.test_relimport import greet_relative

def greet(): print(f"[{os.path.basename(__file__)}] Hello, World!")


if __name__ == "__main__":
    print(f"[{__file__}] __name__ = {__name__}")
    greet()
    greet_relative()
    greet_nested()
    new_addition()



from waivek.timer import Timer
timer = Timer(precision=3)
from typing import TYPE_CHECKING # used in a lot of places, so import it here

from waivek.color import Code
from waivek.common import create_partitions, smart_pad, enumerate_count
from waivek.common import Timestamp, Date
# timer.start("import db_init, insert_dictionaries")
from waivek.db import db_init, insert_dictionaries # slow: 0.03s
# timer.print("import db_init, insert_dictionaries")
from waivek.dbutils import Connection              # slow: 0.02s
from waivek.error2 import handler # still takes 0.02s, improved with TYPE_CHECKING
from waivek.get import aget
from waivek.ic import ic, ib
from waivek.print_utils import head, truncate, abbreviate
from waivek.reltools import rel2abs, read, write, readlines, writelines
# timer.start("import pack, unpack")
from waivek.introspection import pack, unpack
# timer.print("import pack, unpack")
from waivek.log import log, add_file_handler, set_verbose_stdout

# from waivek.data import Countries



from waivek.timer import Timer
timer = Timer(precision=3)

from typing import TYPE_CHECKING # takes 0.01s
if TYPE_CHECKING:
    import pdb

import sys
from typing import Optional, List
from contextlib import contextmanager
from types import FrameType, TracebackType

from waivek.json_decode_error import handle_json_decode_error

# color_arg, color_error_repr, print_error_information {{{
def color_arg(arg):
    BLACK           = '\x1b[30m'
    RED             = '\x1b[31m'
    GREEN           = '\x1b[32m'
    YELLOW          = '\x1b[33m'
    BLUE            = '\x1b[34m'
    MAGENTA         = '\x1b[35m'
    CYAN            = '\x1b[36m'
    WHITE           = '\x1b[37m'
    LIGHTBLACK_EX   = '\x1b[90m'
    LIGHTRED_EX     = '\x1b[91m'
    LIGHTGREEN_EX   = '\x1b[92m'
    LIGHTYELLOW_EX  = '\x1b[93m'
    LIGHTBLUE_EX    = '\x1b[94m'
    LIGHTMAGENTA_EX = '\x1b[95m'
    LIGHTCYAN_EX    = '\x1b[96m'
    LIGHTWHITE_EX   = '\x1b[97m'
    RESET = '\x1b[39m'

    if type(arg) == str:
        arg = YELLOW + repr(arg)
    else:
        arg = CYAN + repr(arg)
    return arg

def color_error_repr(error):
    BLACK           = '\x1b[30m'
    RED             = '\x1b[31m'
    GREEN           = '\x1b[32m'
    YELLOW          = '\x1b[33m'
    BLUE            = '\x1b[34m'
    MAGENTA         = '\x1b[35m'
    CYAN            = '\x1b[36m'
    WHITE           = '\x1b[37m'
    LIGHTBLACK_EX   = '\x1b[90m'
    LIGHTRED_EX     = '\x1b[91m'
    LIGHTGREEN_EX   = '\x1b[92m'
    LIGHTYELLOW_EX  = '\x1b[93m'
    LIGHTBLUE_EX    = '\x1b[94m'
    LIGHTMAGENTA_EX = '\x1b[95m'
    LIGHTCYAN_EX    = '\x1b[96m'
    LIGHTWHITE_EX   = '\x1b[97m'
    BOLD            = '\x1b[1m'
    RESET           = '\x1b[39m'
    RESET_ALL       = '\x1b[0m'
    NORMAL          = '\x1b[22m'

    string = repr(error)
    args = ', '.join(color_arg(arg) for arg in error.args)
    colored_string = BOLD + f"{BOLD + LIGHTRED_EX}{error.__class__.__name__}{NORMAL}({YELLOW}{args}{LIGHTRED_EX}){RESET}" 
    return colored_string
    # lhs = Code.LIGHTRED_EX + m.group(1) 
    # middle =  Code.YELLOW + m.group(2) 
    # # middle = re.sub(r"('\w+')", Code.LIGHTCYAN_EX + r'\1', middle)
    # rhs = Code.LIGHTRED_EX + m.group(3)
    # colored_string = lhs + middle + rhs
    return colored_string

def print_error_information(error: Exception, frame_to_green: FrameType):
    # C:\Users\vivek\Documents\Python                               -> ~/Documents/Python
    # C:\Users\vivek\AppData\Roaming\Python\Python310\site-packages -> $APPDATA/Python/site-packages
    # C:\Program Files\Python310\Lib\site-packages                  -> 
    import traceback
    import os
    from pathlib import Path

    tb = error.__traceback__

    print()
    print(color_error_repr(error))
    # print(Code.LIGHTRED_EX + repr(error))

    from waivek.frame import frame_gen
    from waivek.color import Code
    call_frames = list(frame_gen())
    call_file = call_frames[3].f_code.co_filename

    frames = [ frame for frame, _ in traceback.walk_tb(tb) ]

    summaries = traceback.extract_tb(tb)
    pairs = reversed(list(zip(frames, summaries)))
    
    from waivek.ic import Table
    table = Table()
    table.gutter = '    '
    table.separator = ' ... '

    for i, (frame, summary) in enumerate(pairs):
        filepath = summary.filename
        filepath = str(Path(filepath).resolve()) # C:\users -> C:\Users
        if os.getcwd() in filepath:
            filepath = os.path.relpath(os.path.abspath(filepath))
        else:
            homedir = os.path.expanduser("~")
            filepath = filepath.replace(homedir, "~")

        line_number = summary.lineno
        line = summary.line
        # line = Code.LIGHTGREEN_EX + line if i == 0 else line
        # if green_done is False and frame.f_code.co_filename == call_file:
        if frame == frame_to_green:
            line = Code.LIGHTGREEN_EX + line
            line = '\x1b[1m' + line + '\x1b[22m'

            green_done = True
        lhs_string = f"{filepath}:{line_number}"
        table.row([lhs_string, line])
    table_string = str(table)
    table_lines = table_string.split("\n")
    table_lines = table_lines[1:]
    table_string = "\n".join(table_lines)
    print(table_string)
# }}}

def framestr(frame: FrameType) -> str:
    with open(frame.f_code.co_filename) as f:
        lines = f.readlines()
    line = lines[frame.f_lineno - 1].strip()
    return f'<frame> {frame.f_code.co_filename}:{frame.f_lineno} {line}'

def _select_frame_silent(self: "pdb.Pdb", number: int):
    assert 0 <= number < len(self.stack)
    self.curindex = number
    self.curframe = self.stack[self.curindex][0]
    self.curframe_locals = self.curframe.f_locals
    self.lineno = None

def do_down_silent(self: "pdb.Pdb", arg: int):
    if self.curindex + 1 == len(self.stack):
        print('Newest frame')
        return
    count = arg
    if count < 0:
        newframe = len(self.stack) - 1
    else:
        newframe = min(len(self.stack) - 1, self.curindex + count)
    _select_frame_silent(self, newframe)

def do_up_silent(self: "pdb.Pdb", error: Exception, arg: int):
    if self.curindex == 0:
        print('Oldest frame')
        return
    count = arg
    if count < 0:
        newframe = 0
    else:
        newframe = max(0, self.curindex - count)
    _select_frame_silent(self, newframe)

def x_three(self: "pdb.Pdb", error: Exception, arg):
    """
    Move up the stack by `arg` frames and print the stack frame
    """
    do_up_silent(self, error, arg)
    assert self.curframe
    print_error_information(error, self.curframe)

def make_x_one(self: "pdb.Pdb", error: Exception):
    def x_one(arg):
        if arg == '':
            arg = 1
        else:
            arg = int(arg)
        x_three(self, error, arg)
    return x_one

def y_three(self: "pdb.Pdb", error: Exception, arg):
    """
    Move down the stack by `arg` frames and print the stack frame
    """
    do_down_silent(self, arg)
    assert self.curframe
    print_error_information(error, self.curframe)

def make_y_one(self: "pdb.Pdb", error: Exception):
    def y_one(arg):
        if arg == '':
            arg = 1
        else:
            arg = int(arg)
        y_three(self, error, arg)
    return y_one

def make_v_zero_arg(self: "pdb.Pdb", error: Exception):
    """
    Print the variables in the current stack frame
    """
    from waivek.error import print_variables
    def v_zero_arg(*args):
        assert self.curframe
        print_variables(self.curframe.f_locals)
    return v_zero_arg

def make_z_zero_arg(self: "pdb.Pdb", error: Exception):
    """
    Print the stack frame
    """
    def z_zero_arg(*args):
        assert self.curframe
        print_error_information(error, self.curframe)
    return z_zero_arg

def ic_two(self: "pdb.Pdb", arg):
    from waivek.ic import ic
    assert self.curframe
    val = eval(arg, self.curframe.f_globals, self.curframe.f_locals)
    ic(val)

def make_ic_one_arg(self: "pdb.Pdb"):
    def ic_one(arg):
        ic_two(self, arg)
    return ic_one

def ib_two(self: "pdb.Pdb", arg):
    from waivek.ic import ib
    assert self.curframe
    val = eval(arg, self.curframe.f_globals, self.curframe.f_locals)
    ib(val)

def make_ib_one_arg(self: "pdb.Pdb"):
    def ib_one(arg):
        ib_two(self, arg)
    return ib_one

def print_error_information_from_pdb(self: "pdb.Pdb", error: Exception):
    # frame = self.curframe
    # assert frame
    if not self.curframe:
        raise ValueError('No frame selected')
    print_error_information(error, self.curframe)

@contextmanager
def handler():
    from json import JSONDecodeError # takes 0.01s - 0.02s
    import traceback # takes 0.002s, not much but still inlining
    try:
        yield
    except Exception as error:
        if isinstance(error, JSONDecodeError):
            handle_json_decode_error(error)
        if type(error).__name__ == 'bdb.BdbQuit':
            pass
        else:
            # pdb.post_mortem(error.__traceback__); return
            import pdb

            caller_file = sys._getframe(2).f_code.co_filename

            Pdb = pdb.Pdb()
            Pdb.reset()

            # CUSTOMIZATION 2: Move the debugger to the frame where the error occurred within the file from which the program was run, instead of the frame where the error was raised
            frames = [ frame for frame, _ in traceback.walk_tb(error.__traceback__) ]
            # needle_frame = next(frame for frame in reversed(frames) if frame.f_code.co_filename == __file__)
            needle_frame = next(frame for frame in reversed(frames) if frame.f_code.co_filename == caller_file)
            Pdb.setup(f=None, tb=error.__traceback__) # NOTE[1/1]: Do not do `Pdb.setup(f=needle_frame, tb=error.__traceback__)` as it creates new frames or something

            _select_frame_silent(Pdb, frames.index(needle_frame))

            # CUSTOMIZATION 3: Add `x` and `y` commands for moving up and down the stack in a customized way, and `z` command for printing variables
            Pdb.do_x = make_x_one(Pdb, error)      # pyright:ignore[reportAttributeAccessIssue]
            Pdb.do_y = make_y_one(Pdb, error)      # pyright:ignore[reportAttributeAccessIssue]
            Pdb.do_z = make_z_zero_arg(Pdb, error) # pyright:ignore[reportAttributeAccessIssue]
            Pdb.do_v = make_v_zero_arg(Pdb, error) # pyright:ignore[reportAttributeAccessIssue]

            # CUSTOMIZATION 1: Add `ic` and `ib` aliases for exploration
            Pdb.do_ic = make_ic_one_arg(Pdb)       # pyright:ignore[reportAttributeAccessIssue]
            Pdb.do_ib = make_ib_one_arg(Pdb)       # pyright:ignore[reportAttributeAccessIssue]

            # CUSTOMIZATION 4: Nicer stacktrace
            print_error_information_from_pdb(Pdb, error)
            Pdb._cmdloop()

def select_frame_for_file(frames: List[FrameType], target_file: str) -> Optional[FrameType]:
    for frame in reversed(frames):
        if frame.f_code.co_filename == target_file:
            return frame
    return None

def raise_simple_json_decode_error():
    import json
    string = ''
    json.loads(string)
    raise json.JSONDecodeError('Expecting value', '', 0)

def main():
    # files:
    #   [1] json_decode_error.py

    raise_simple_json_decode_error()
    # raise Exception('This is an error', 1)

if __name__ == "__main__":
    with handler():
        main()
# run.vim: vert term python waivek/__init__.py



import os
import sys
from waivek.truncate import get_display_length, truncate
from waivek.common import enumerate_count

from color import Code
import rich

def is_tuple_table(table):
    if type(table) != list:
        return False
    types = list(set(type(row) for row in table))
    if len(types) != 1:
        return False
    common_type = types[0]
    return common_type in [ tuple, list ]

def is_single_dict_table(table):
    # we add items check so that Date class doesn’t pass
    return type(table) == dict or (hasattr(table, '__dict__') and hasattr(table, 'items'))

def is_multi_dict_table(table):
    if type(table) != list:
        return False
    if any(type(row) != dict for row in table):
        return False
    from itertools import groupby
    unique_key_combinations = list(groupby([ D.keys() for D in table ]))
    return len(unique_key_combinations) == 1

def is_db_table(table):
    if type(table) != list:
        return False
    from waivek.dbutils import get_sqlite3
    sqlite3 = get_sqlite3()
    boolean = all(type(row) == sqlite3.Row for row in table) 
    return boolean

def get_max_width_and_height():
    in_terminal = sys.stdout.isatty()
    if in_terminal:
        return os.get_terminal_size()
    else:
        return (200, 200)
def ljust_display(s, width):
    fill = " "
    display_length = get_display_length(s)
    if display_length < width:
        return s + fill * (width - display_length)
    else:
        return s

def rjust_display(s, width):
    fill = " "
    display_length = get_display_length(s)
    if display_length < width:
        return fill * (width - display_length) + s
    else:
        return s

def pad_lines(lines: list[str], line_count=1):
    """
    Adjusts the input list of lines to match the specified line_count.

    If the number of input lines is less than or equal to line_count,
    the function pads the list with blank lines to reach the desired height.
    If the number of input lines exceeds line_count, it truncates the list
    from both ends and inserts an ellipsis ("...") in the middle.
    """
    list_height = len(lines)
    if list_height <= line_count:
        # vjust
        blank_line_count = line_count - list_height
        max_length = max(len(line) for line in lines)
        blank_lines = [ " " * max_length for _ in range(blank_line_count) ]
        return lines + blank_lines
    else:
        # vtruncate
        top_end = int(line_count/2)
        bot_start = list_height - int(line_count/2) + 1
        truncated_lines = lines[0:top_end] + [ "..." ] + lines[bot_start:]
        # if len(slice_L) != line_count: error {{{
        if len(truncated_lines) != line_count:
            print(Code.RED +  "(error) v_pad(list_string, line_count)")
            print(Code.RED + f"....... len(slice_L): {len(truncated_lines)}")
            print(Code.RED + f"....... line_count  : {line_count}")
            raise Exception("len(slice_L) != line_count")
        # }}}
        return truncated_lines

class Table:

    # Why we can’t use columnar
    #
    # 1. No option for separator only between columns. Any separator you choose is also part of the border
    # 2. 150ms import time vanilla 

    def __init__(self):
        self.gutter = '  '
        self.separator = '  '
        self.table = []

    def row(self, *args):
        if len(args) == 1:
            L = args[0]
        else:
            L = list(args)
        self.table.append(L)

    def parse(self, D):
        self.table = D

    def normalize_table(self):
        table = self.table
        if is_tuple_table(table):
            L = [ tuple(sublist) for sublist in table ]
            return [], L
        elif is_single_dict_table(table):
            assert hasattr(table, 'items')
            items = list(table.items()) # pyright:ignore[reportAttributeAccessIssue]
            return [ "KEY", "VALUE" ], items
        elif is_db_table(table): # Transform to multi_dict_table, DRY Violated
            multi_dict_table = [ dict(row) for row in table ]
            keys = multi_dict_table[0].keys()
            L = [ tuple(D[key] for key in keys) for D in multi_dict_table ]
            return keys, L
        elif is_multi_dict_table(table):
            keys = table[0].keys()
            L = [ tuple(D[key] for key in keys) for D in table ]
            return keys, L

        return None

    def get_column_tuples(self, headers, tuple_table):
        def get_column_just(column):
            is_numeric = all(type(cell) in [int, float] for cell in column)
            return rjust_display if is_numeric else ljust_display
        def get_width(cell):
            # Handles multi-line strings
            cell_str = str(cell)
            val = max(get_display_length(line) for line in cell_str.split("\n")) # get_display_length
            return val
        def distribute(initial_lengths, terminal_length):
            final_lengths = [ 0 for _ in initial_lengths ]
            total_length = 0
            while total_length != terminal_length:
                all_lengths_satisfied = all([ initial_length == final_length for initial_length, final_length in zip(initial_lengths, final_lengths) ])
                if all_lengths_satisfied:
                    break
                for i, length in enumerate(initial_lengths):
                    if final_lengths[i] == initial_lengths[i]:
                        continue
                    final_lengths[i] = final_lengths[i] + 1
                    total_length += 1
                    if total_length == terminal_length:
                        break
            return final_lengths

        column_count = len(tuple_table[0])
        columns = [ list(row[i] for row in tuple_table) for i in range(column_count) ]
        column_tuples = [ ( max(get_width(cell) for cell in column), get_column_just(column) ) for column in columns  ]
        if headers:
            column_tuples = [ (max(width, len(header)), just) for header, (width, just) in zip(headers, column_tuples) ]

        # Distribute Column Widths given terminal size
        initial_widths = [ width for width, _ in column_tuples ]
        
        terminal_width, _ = get_max_width_and_height()
        gutter_width = len(self.gutter) * 2
        separator_width = len(self.separator) * (column_count-1)
        column_width_total = terminal_width - gutter_width - separator_width
        final_widths = distribute(initial_widths, column_width_total)
        for i, (width, just) in enumerate(column_tuples):
            column_tuples[i] = (final_widths[i], just)
        return column_tuples

    def to_string(self):

        def fmt(cells, column_tuples):
            S = self.separator
            G = self.gutter
            all_strings_are_single_line = all("\n" not in str(cell) for cell in cells)
            if all_strings_are_single_line:
                return G + S.join(just(truncate(str(cell), width), width) for cell, (width, just) in zip(cells, column_tuples))
            else:
                cells = [ str(cell) for cell in cells ]
                largest_height = max(len(cell.split("\n")) for cell in cells)
                v_pad_cells = [ pad_lines(cell.split("\n"), largest_height) for cell in cells ]
                return_lines = []
                for i in range(largest_height):
                    parts = [ cell[i] for cell in v_pad_cells ]
                    parts = [ just(truncate(part, width), width) for part, (width, just) in zip(parts, column_tuples) ] 
                    return_lines.append(G + S.join(parts))
                return "\n".join(return_lines)

        result = self.normalize_table()
        assert result
        headers, tuple_table = result
        column_tuples = self.get_column_tuples(headers, tuple_table)

        header_lines = [ "", fmt(headers, column_tuples), "" ] if headers else [ "" ]
        footer_lines = [ "" ]
        row_lines = [ fmt(row, column_tuples) for row in tuple_table ]
        table_string = "\n".join(header_lines+row_lines+footer_lines)
        return table_string

    def __str__(self):
        return self.to_string()

def to_table(rows):
    T = Table()
    for row in rows:
        T.row(row)
    return T

def test_table():
    from waivek.ic import ic
    merged_test_cases = [
        ("is_tuple_table", [ ("a", "b", "c"), ("d", "e", "f"), ("g", "h", "i") ]),
        ("is_tuple_table", [(1, 2), (3, 4)]),   # List of tuples
        ("is_tuple_table", [[1, 2], [3, 4]]),   # List of lists

        ("is_single_dict_table", {"d": "e", "a": "b"}),  # Dictionary
        ("is_single_dict_table", [{"key": "value"}]), # List of dictionaries

        ("is_multi_dict_table", [{"key1": "value1"}, {"key1": "value2"}]),  # Same keys

    ]

    ic({"d": "e", "a": "b"})
    name, table = merged_test_cases[0]
    for count_str, (name, table) in enumerate_count(merged_test_cases):
        try:
            T = to_table(table)
            print(T)
        except Exception as error:
            rich.print(f"{count_str} [black on red] FAIL [/] {repr(table)}")
            print(error.__class__.__name__)
            print("===")
        rich.print(f"{count_str} [black on green] PASS [/]")
        print("===")

if __name__ == "__main__":
    test_table()




# https://www.python.org/dev/peps/pep-0508/#environment-markers
from setuptools import setup
import sys
import importlib.util
# check if module rich is installed

# import rich
import pip._vendor.rich as rich

def get_new_version():
    import urllib.request
    from urllib.error import URLError
    import json
    package_name = "waivek"
    url = f"https://pypi.org/pypi/{package_name}/json"
    try:
        req = urllib.request.urlopen(url)
    # except urllib.request.URL
    # catch URLError
    except URLError as e:
        error = e
        print(f"Error: {error}")
        print(f"Could not fetch {url}")
        sys.exit(1)

    D = json.load(req)
    version = D["info"]["version"]
    major, minor, patch = version.split(".")
    patch = int(patch) + 1
    new_version = f"{major}.{minor}.{patch}"
    print(f"{package_name} version: {version} -> {new_version}")
    return new_version

def remove_dist_directory():
    import shutil
    shutil.rmtree("dist", ignore_errors=True)

def get_usage():
    if len(sys.argv) >= 3 and sys.argv[1] in ['egg_info', 'install', 'clean']:
        # `pip install .` -> ['/home/vivek/python_common/setup.py', 'egg_info', '--egg-base', '/tmp/pip-pip-egg-info-knsmhso3']
        return "PIP_INSTALL_DOT"
    # ['/home/vivek/python_common/setup.py', 'bdist_wheel', '-d', '/tmp/pip-wheel-4pt1wlji']
    if len(sys.argv) >= 3 and sys.argv[1] == 'bdist_wheel' and sys.argv[2] == '-d':
        return "PIP_INSTALL_DOT"
    if len(sys.argv) >= 3 and sys.argv[1] == 'sdist' and sys.argv[2] == 'bdist_wheel':
        return "UPLOAD_TO_PYPI"
    return "UNKNOWN"

temp_usage = get_usage()
if temp_usage == "UNKNOWN":
    usage_message = r"""
    Passd args: {0}

    [red]Usage:[/red] 

        [bold]python setup.py sdist bdist_wheel[/bold]

    Then upload the package to PyPI using:

        [bold]twine upload dist/*[/bold]

    If you get asked for a username and password, you need to have [bold]~/.pypirc[/bold] file with the following content:

        \[pypi]
        username: __token__ 
        password: <your api token, with `pypi` prefix>
    """.format(sys.argv)
    rich.print(usage_message)
    sys.exit(1)

usage = get_usage()

if usage == "UPLOAD_TO_PYPI":
    missing_modules = False
    if importlib.util.find_spec("twine") is None:
        rich.print("Please install twine using [bold]pip install twine[/bold]")
        missing_modules = True
    if importlib.util.find_spec("wheel") is None:
        rich.print("Please install wheel using [bold]pip install wheel[/bold]")
        missing_modules = True
    if missing_modules:
        sys.exit(1)

if usage == "UPLOAD_TO_PYPI":
    remove_dist_directory()

new_version = "0.0.0" 
if usage == "UPLOAD_TO_PYPI":
    new_version = get_new_version()
py_modules = ['color', 'common', 'data', 'db', 'error', 'frame', 'get', 'ic', 'print_utils', 'reltools', 'timer', 'trace']
py_modules = ['waivek.' + x for x in py_modules]
long_description = open('README.md').read()
setup(
        name='waivek',
        version=new_version,
        packages=['waivek'],
        py_modules = py_modules,
        install_requires = [
            'aiohttp>=3.8.1',
            'executing==0.8.2',
            'python_dateutil==2.8.2',
            'loguru',# for `log.py`
            'timeago==1.0.14',
            'columnar==1.3.1',
            'pysqlite3-binary; platform_system=="Linux"'
        ],
        long_description=long_description,
        long_description_content_type='text/markdown'
)




import re

def truncate(s, max_length):
    ansi_escape = re.compile(r'\x1b\[[0-9;]*[a-zA-Z]')
    
    # Tokenize the string into (text, is_visible, styles) tuples
    tokens = []
    pos = 0
    active_styles = []
    while pos < len(s):
        m = ansi_escape.match(s, pos)
        if m:
            code = m.group()
            if code == '\x1b[0m':
                # Reset code clears all styles
                active_styles.clear()
            else:
                # Add style to active styles
                active_styles.append(code)
            tokens.append((code, False, code))
            pos = m.end()
        else:
            # Visible character
            tokens.append((s[pos], True, ''.join(active_styles)))
            pos += 1

    # Calculate the visible length (excluding ANSI codes)
    visible_length = sum(1 for t, is_visible, _ in tokens if is_visible)

    # If the visible length is within the maximum, return the original string
    if visible_length <= max_length:
        return s

    separator = "..."
    sep_len = len(separator)
    visible_chars_to_keep = max_length - sep_len

    if visible_chars_to_keep <= 0:
        # Not enough space even for the separator and any chars
        return separator[:max_length]

    # **Prioritize the right side (end of the string)**
    # Split the visible characters between the start and end
    start_visible_chars = visible_chars_to_keep // 2
    end_visible_chars = visible_chars_to_keep // 2 + visible_chars_to_keep % 2

    # Collect tokens from the start
    start_tokens = []
    visible_count = 0
    i = 0
    while i < len(tokens):
        t, is_visible, styles = tokens[i]
        start_tokens.append((t, is_visible, styles))
        if is_visible:
            visible_count += 1
            if visible_count >= start_visible_chars:
                # Include any subsequent ANSI codes
                i += 1
                while i < len(tokens) and not tokens[i][1]:
                    start_tokens.append(tokens[i])
                    i += 1
                break
        i += 1
    # Save the active styles at the truncation point
    active_styles_start = tokens[i-1][2] if i > 0 else ''

    # Collect tokens from the end
    end_tokens = []
    visible_count = 0
    i = len(tokens) - 1
    while i >= 0:
        t, is_visible, styles = tokens[i]
        end_tokens.insert(0, (t, is_visible, styles))
        if is_visible:
            visible_count += 1
            if visible_count >= end_visible_chars:
                # Include any preceding ANSI codes
                i -= 1
                while i >= 0 and not tokens[i][1]:
                    end_tokens.insert(0, tokens[i])
                    i -= 1
                break
        i -= 1
    # Save the active styles at the start of the end tokens
    active_styles_end = end_tokens[0][2] if end_tokens else ''

    # Close any open styles before the separator
    reset_code = '\x1b[0m'
    if active_styles_start and not start_tokens[-1][0].endswith(reset_code):
        start_tokens.append((reset_code, False, ''))

    # Re-open styles after the separator if needed
    if active_styles_end:
        end_tokens.insert(0, (active_styles_end, False, active_styles_end))

    # Combine the tokens with the separator
    truncated_tokens = start_tokens + [(separator, True, '')] + end_tokens

    # Reconstruct the string
    result = ''.join(t for t, is_visible, _ in truncated_tokens)
    return result

# run.vim: vert term python test_truncate.py



import unittest
from unittest.mock import Mock, patch, create_autospec
from types import FrameType, CodeType, TracebackType
from typing import List
import pdb
from json.decoder import JSONDecodeError
# Import the functions you want to test
from waivek.error2 import (
    select_frame_for_file,
    _select_frame_silent,
    do_down_silent,
    do_up_silent,
    handler,
)

def create_mock_frame(filename: str) -> FrameType:
    mock_code = create_autospec(CodeType, spec_set=True)
    mock_code.co_filename = filename
    mock_frame = create_autospec(FrameType, spec_set=True)
    mock_frame.f_code = mock_code
    return mock_frame

class TestDebugger(unittest.TestCase):
    def test_select_frame_for_file(self):
        mock_frames: List[FrameType] = [
            create_mock_frame("file1.py"),
            create_mock_frame("file2.py"),
            create_mock_frame("file3.py"),
        ]
        result = select_frame_for_file(mock_frames, "file2.py")
        self.assertEqual(result, mock_frames[1], "Should select the correct frame")
        result = select_frame_for_file(mock_frames, "nonexistent.py")
        self.assertIsNone(result, "Should return None for non-existent file")

    def test_select_frame_silent(self):
        mock_pdb = Mock(spec=pdb.Pdb)
        mock_pdb.stack = [(create_mock_frame(f"file{i}.py"), i) for i in range(3)]
        _select_frame_silent(mock_pdb, 1)
        self.assertEqual(mock_pdb.curindex, 1, "Should set correct curindex")
        self.assertEqual(mock_pdb.curframe, mock_pdb.stack[1][0], "Should set correct curframe")
        self.assertEqual(mock_pdb.curframe_locals, mock_pdb.curframe.f_locals, "Should set correct curframe_locals")
        self.assertIsNone(mock_pdb.lineno, "Should set lineno to None")

    def test_do_down_silent(self):
        mock_pdb = Mock(spec=pdb.Pdb)
        mock_pdb.stack = [(create_mock_frame(f"file{i}.py"), i) for i in range(3)]
        mock_pdb.curindex = 0
        do_down_silent(mock_pdb, 1)
        self.assertEqual(mock_pdb.curindex, 1, "Should move down one frame")

    def test_do_up_silent(self):
        mock_error = Mock(spec=Exception)
        mock_pdb = Mock(spec=pdb.Pdb)
        mock_pdb.stack = [(create_mock_frame(f"file{i}.py"), i) for i in range(3)]
        mock_pdb.curindex = 2
        do_up_silent(mock_pdb, mock_error, 1)
        self.assertEqual(mock_pdb.curindex, 1, "Should move up one frame")

    @patch('pdb.Pdb')
    @patch('traceback.walk_tb')
    def test_handler(self, mock_walk_tb, mock_pdb_class):
        mock_pdb = Mock(spec=pdb.Pdb)
        mock_pdb_class.return_value = mock_pdb
        mock_error = JSONDecodeError('Expecting value', '', 0)
        mock_error.__traceback__ = create_autospec(TracebackType, spec_set=True)
        mock_walk_tb.return_value = [
            (create_mock_frame("file1.py"), None),
            (create_mock_frame(__file__), None),
        ]
        with patch('builtins.__import__', side_effect=mock_error):
            with handler():
                __import__('json')  # This will raise the JSONDecodeError
        mock_pdb.reset.assert_called_once()
        mock_pdb.setup.assert_called_once()
        self.assertTrue(hasattr(mock_pdb, 'do_x'), "Should add custom 'x' command")
        self.assertTrue(hasattr(mock_pdb, 'do_y'), "Should add custom 'y' command")
        self.assertTrue(hasattr(mock_pdb, 'do_z'), "Should add custom 'z' command")
        mock_pdb._cmdloop.assert_called_once()

if __name__ == '__main__':
    unittest.main()
